{-# LANGUAGE GADTs #-}
module Util where

import Control.Arrow
import System.IO.Unsafe
import System.IO

import Data.IntMap (IntMap)
import qualified Data.IntMap as IntMap
import Data.Monoid

newtype MIntMap a = MIntMap { fromMIntMap :: IntMap a }

instance Monoid a => Monoid (MIntMap a) where
  mempty = MIntMap mempty
  (MIntMap x) `mappend` (MIntMap y) = MIntMap (IntMap.unionWith mappend x y)

mIntMapMod :: (IntMap a -> IntMap a) -> MIntMap a -> MIntMap a
mIntMapMod f (MIntMap x) = MIntMap (f x)

eachWithContext :: [a] -> [(a, [a])]
eachWithContext []      = []
eachWithContext (x:xs)  = (x, xs):map (second (x:)) (eachWithContext xs)

head' :: String -> [a] -> a
head' _ (x:_)  = x
head' errMsg _ = error errMsg

trace :: String -> a -> a
trace str x = unsafePerformIO $ do
  putStrLn str
  hFlush stdout
  return x

mTrace :: (Monad m) => String -> m ()
mTrace s = trace s (return ())

fst3 :: (a,b,c) -> a
fst3 (a,_,_) = a

snd3 :: (a,b,c) -> b
snd3 (_,b,_) = b

first3 :: (a -> b) -> (a, x, y) -> (b, x, y)
first3 f (a, x, y) = (f a, x, y)

zipWithM3 :: (Monad m) => (a -> b -> c -> m d) -> [a] -> [b] -> [c] -> m [d]
zipWithM3 f x y = sequence . zipWith3 f x y

fromLeft :: String -> Either a b -> a
fromLeft msg (Right _) = error ("fromLeft: " ++ msg)
fromLeft _   (Left v)  = v

data ExistsClass a c where
  ExistsC :: c b -> a b -> ExistsClass a c
