{-# LANGUAGE ScopedTypeVariables #-}
module Util.Graph
( minimizeGraph
) where

import Data.Graph.Inductive
import Data.List
import Data.Ord
import Data.Function

minimizeGraph :: (Ord a, Ord b) => Gr a b -> Gr a b
minimizeGraph inGraph = loop inGraph

loop :: (Ord a, Ord b) => Gr a b -> Gr a b
loop gr
  | nextStep `equal` gr  = gr
  | otherwise            = loop nextStep
  where
    nextStep = oneStep gr

oneStep :: forall a b . (Ord a, Ord b) => Gr a b -> Gr a b
oneStep inGraph = outGraph
  where
    outGraph = foldl' combine inGraph equivGroups 
      where
        combine gr xs = case filter (flip gelem gr . fst) xs of
          []     -> gr
          (n:ns) -> calcStep n ns gr

    equivGroups :: [[LNode a]]
    equivGroups = groupBy ((==) `on` check) . sortBy (comparing check) $ labNodes inGraph
      where
        check :: LNode a -> (a, [(Node, b)])
        check n = (snd n, sort . lsuc inGraph . fst $ n)

    calcStep :: LNode a -> [LNode a] -> Gr a b -> Gr a b
    calcStep (mn, _) equivNodes inGraph = outGraph
      where
        equivEdges :: [(Node, b)]
        equivEdges = concatMap (lpre inGraph . fst) 
                   $ equivNodes

        newEdges :: [LEdge b]
        newEdges = map (\(n, l) -> (n, mn, l)) equivEdges

        outGraph = insEdges newEdges 
                 . delNodes (map fst equivNodes) 
                 $ inGraph
