{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE NamedFieldPuns #-}
module Vis.Rules where

import Data.IORef
import Engine.EngineM
import Engine.ExperimentResult
import Prelude hiding (exp)
import GHC
import GraphRewriting
import HscTypes
import Plan
import Plan.TreePlan
import qualified Data.Foldable as F
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import System.IO.Unsafe
import Vis.Graph
import Engine.Experiment
import Data.Maybe

data Env = Env { session :: Session 
               , engineState :: EngineState
               , engineEnv :: EngineEnv
               }

runEngineRef :: IORef Env -> EngineM a -> a
runEngineRef ref code = unsafePerformIO $ do
  env@Env { session , engineState , engineEnv } <- readIORef ref
  (a, s', _) <- flip reflectGhc session $ runEngineM engineState engineEnv code
  writeIORef ref env { engineState = s' }
  return a

planToDo :: (View [Port] n, View PlanNode n) => IORef Env -> Rule n
planToDo envRef = do
  PlanToDo { inp , plan } <- node
  case plan of TPStep { tpExperiment , tpOnOk , tpOnDataCon , tpOnBottom } -> replace 1 $ 
                \[newEdge] -> runEngineRef envRef $ do 
                  (getExperimentResult -> expRes) <- runExperiment tpExperiment 
                  let nextNode = case expRes of
                                  Redundant _ -> Nil newEdge
                                  Ok          -> PlanToDo { inp  = newEdge
                                                          , plan = tpOnOk     -- TODO: tpOnDataCon!
                                                          }
                                  Errors _    -> Nil newEdge
                                  Bottoms id  -> PlanToDo { inp = newEdge
                                                          , plan = fromJust $ Map.lookup id tpOnBottom
                                                          }
                                  Property _  -> PlanToDo { inp = newEdge
                                                          , plan = tpOnOk
                                                          }
                  return [Node $ RanExperiment { inp = inp
                                               , exp = tpExperiment
                                               , result = expRes
                                               , out = newEdge
                                               }
                         , Node $ nextNode 
                         ]
                  
               TPSplit { tpChildren } -> replace (Seq.length tpChildren) $ 
                \kidEdges -> map Node $ Split { inp = inp , children = kidEdges } : zipWith PlanToDo kidEdges (F.toList tpChildren)
