{-# LANGUAGE MagicHash #-}
module Engine.ExperimentCacheTests where

import Control.Monad.State.Strict
import Control.Monad.Trans
import Data.Maybe
import Engine.EngineHValue
import Engine.ExperimentCache
import Engine.ExperimentCacheTestsLib
import Engine.ExperimentResult
import GHC.Prim
import System.Mem.StableName
import Test.HUnit ((~:))
import TysWiredIn

tests = "ExperimentCacheTests" ~: mkTests test

test 0 = with (fifoStrategy 1) x { e = 2 } $ \Args { es = [e1, e2] } -> do
  apply $ storeExperimentResult e1 Ok
  Just Ok <- check $ lookupExperimentResult e1
  Nothing <- check $ lookupExperimentResult e2
  done


test 1 = with (fifoStrategy 1) x { e = 3 } $ \Args { es = [e1, e2, e3] } -> do
  apply $ storeExperimentResult e2 Ok
        . storeExperimentResult e1 Ok
  Nothing <- check $ lookupExperimentResult e1
  Just Ok <- check $ lookupExperimentResult e2
  Nothing <- check $ lookupExperimentResult e3
  done


test 2  = with (fifoStrategy 2) x { e = 3 } $ \Args { es = [e1, e2, e3] } -> do
  Nothing <- check $ lookupExperimentResult e1

  apply $ storeExperimentResult e1 Ok
  Just Ok <- check $ lookupExperimentResult e1
  Nothing <- check $ lookupExperimentResult e2

  apply $ storeExperimentResult e2 Ok
  Just Ok <- check $ lookupExperimentResult e1
  Just Ok <- check $ lookupExperimentResult e2
  Nothing <- check $ lookupExperimentResult e3

  apply $ storeExperimentResult e3 Ok
  Nothing <- check $ lookupExperimentResult e1
  Just Ok <- check $ lookupExperimentResult e2
  Just Ok <- check $ lookupExperimentResult e3

  apply $ storeExperimentResult e1 (Property False)
  Just (Property False) <- check $ lookupExperimentResult e1
  Nothing <- check $ lookupExperimentResult e2
  Just Ok <- check $ lookupExperimentResult e3
  done


test 3 = with (fifoStrategy 1) x { e = 2, h = 2, i = 2 } $ \Args { es = [e1, e2]
                                                                 , hs = [h1, h2]
                                                                 , is = [i1, i2] } -> do
  Nothing <- check $ lookupExperimentResult e1

  apply $ storeEngineHValue e1 h1

  Just exps <- check $ lookupHValueExperiments h1 
  checkEqExps [e1] exps
  Just Ok <- check $ lookupExperimentResult e1

  apply $ storeEngineHValue e2 h2
  Nothing <- check $ lookupHValueExperiments h1
  Just exps <- check $ lookupHValueExperiments h2
  checkEqExps [e2] exps
  Nothing <- check $ lookupExperimentResult e1
  Just Ok <- check $ lookupExperimentResult e2
  done

test 4 = with (fifoStrategy 2) x { e = 3, h = 2, i = 2 } $ \Args { es = [e1, e2, e3]
                                                                 , hs = [h1, h2]
                                                                 , is = [i1, i2]
                                                                 }-> do 
  apply $ storeEngineHValue e1 h1
  apply $ storeEngineHValue e2 h1

  Just exps <- check $ lookupHValueExperiments h1
  checkEqExps [e1, e2] exps

  apply $ storeEngineHValue e3 h2

  Just exps <- check $ lookupHValueExperiments h1
  checkEqExps [e2] exps

test _ = notWrittenYet
