{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE MagicHash #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
module Engine.ExperimentCacheTestsLib
( done
, mkTests
, apply
, check
, checkEqExps
, checkEqId
, x
, with
, Args(..)
, ArgConfig (e, h, i)
, notWrittenYet
) where

import Control.Monad
import Control.Monad.State.Strict
import GHC.Paths
import Data.Maybe
import Data.Sequence (Seq)
import Engine.EngineHValue
import Engine.ExperimentCache
import FastString
import Format
import GHC
import GHC.Prim
import Id
import Outputable
import Plan
import qualified Data.Foldable as F
import qualified Data.Sequence as Seq
import Test.HUnit hiding (State)
import TysWiredIn
import Unique

newtype ECT a = ECT { runECT :: StateT ExperimentCache IO a }
  deriving Monad

data Args 
  = Args 
  { es :: [Experiment]
  , hs :: [EngineHValue]
  , is :: [Id]
  }

data ArgConfig 
  = ArgConfig
  { e :: Int
  , h :: Int
  , i :: Int
  }

done :: ECT ()
done  = return ()

mkTests :: (Int -> Maybe (IO ())) -> [Test]
mkTests f = map TestCase . catMaybes . takeWhile isJust . map (f$) $ [0..]

apply :: (ExperimentCache -> ExperimentCache) -> ECT ()
apply fn = ECT $ modify fn

check :: (ExperimentCache -> a) -> ECT a
check fn = ECT $ gets fn

checkEqExps :: [Experiment] -> Seq Experiment -> ECT ()
checkEqExps e1s e2s = ECT . liftIO $ e1s == (F.toList e2s) @? (format e1s ++ " != " ++ format e2s)

checkEqId :: Id -> Id -> ECT ()
checkEqId i1 i2 = ECT . liftIO $ i1 == i2 @? (showPpr i1 ++ " != " ++ showPpr i2)

x :: ArgConfig
x = ArgConfig { e = 0, h = 0, i = 0 }

with :: Strategy -> ArgConfig -> (Args -> ECT ()) -> Maybe (IO ())
with strategy config assertionF = Just $ do
  runGhc Nothing . liftIO $ do
    let initCache = newCache strategy

    let experiments = map  mkExperiment [1..e config]
    hvalues        <- mapM mkHValue     [1..h config]
    let ids         = map  mkId         [1..i config]

    let args = Args { es = experiments
                    , hs = hvalues
                    , is = ids
                    }

    flip evalStateT initCache . runECT . assertionF $ args
    

mkExperiment :: Int -> Experiment
mkExperiment i = Experiment { experimentRoot        = IdRoot expId
                            , experimentArgs        = Seq.empty
                            , experimentIsProperty  = False
                            , experimentHasBottom   = False
                            , experimentType        = idType expId
                            , experimentMaxDepth    = -1
                            , experimentSize        = -1
                            }
  where
    expId = mkId i
  

mkId :: Int ->  Id
mkId x = mkSysLocal (mkFastString $ "ectl-mkId-" ++ show x) (mkUnique 'i' x) ty
  where
    ty
      | even x = intTy
      | odd x  = floatTy

mkHValue :: Int -> IO EngineHValue
mkHValue !x 
  | even x  = mkEngineHValue (unsafeCoerce# x)  intTy
  | odd x   = mkEngineHValue (unsafeCoerce# x') floatTy
  where
    x' :: Float
    x' = fromIntegral x

notWrittenYet :: Maybe (IO ())
notWrittenYet = Nothing
