{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE TupleSections #-}
{-# LANGUAGE FlexibleInstances #-}
module IdDB.ConstructorGraphTestLib
( (~~>)
, (==>)
, (!)
, buildTests
, gives
, n
, notWrittenYet
, with
, with_limit
, t 
) where

import Control.Applicative
import Control.Arrow
import Control.Monad
import Data.IntMap (IntMap)
import Data.List
import Data.Maybe
import Data.Monoid
import GHC
import GHC.Paths
import IdDB.ConstructorGraph (ConstructorGraph, DestructPlan)
import IdDB.ConstructorGraphTestUtil
import MonadUtils
import qualified Data.IntMap as IntMap
import qualified IdDB.ConstructorGraph as CG
import IdDB.ConstructorGraphDebug
import qualified Language.Haskell.TH as TH
import qualified Test.HUnit as HUnit
import qualified Util.UniqMap as UniqMap
import qualified Util.UniqSet as UniqSet
import Test.HUnit ((~:), (@?))
import TysWiredIn
import Util.TH
import Util.UniqMap (UniqMap)
import Util.UniqSet (UniqSet)

data TestId 
  = TestId 
  { testIdTypeQ :: TH.TypeQ 
  , testIdId    :: Int
  }

instance Show TestId where
  show (TestId _ x) = "TestId#" ++ show x

data CGExpected
  = CGE (TH.Name, Int) CGRHS
  deriving Show

data CGRHS
  = StopsHere
  | CarrysOn
  | Multiple [CGExpected]
  deriving Show

data CGAssertion = CGA

class CGExpectedLike a where
  toCGRHS :: a -> CGRHS

instance CGExpectedLike CGRHS where
  toCGRHS = id

instance CGExpectedLike [CGExpected] where
  toCGRHS = Multiple

(~~>) :: (CGExpectedLike a) => (TH.Name, Int) -> a -> CGExpected
tidCtr ~~> x = CGE tidCtr (toCGRHS x)

buildTests :: (Int -> Maybe HUnit.Test) -> HUnit.Test
buildTests f = HUnit.TestList $ catMaybes . takeWhile isJust . map (f$) $ [0..]

with :: [ TestId ] -> [(ConstructorGraph TestId -> Ghc ())] -> Maybe HUnit.Test
with = with_limit 1000

with_limit :: Integer -> [ TestId ] -> [(ConstructorGraph TestId -> Ghc ())] -> Maybe HUnit.Test
with_limit limit testIds mkAssertions =  Just $ HUnit.TestCase $ runGhc (Just libdir) $ do
  dflags <- getSessionDynFlags 
  setSessionDynFlags dflags
  defaultErrorHandler dflags $ do
    defaultCleanupHandler dflags $ do
      t1 <- guessTarget ("IdDB.ConstructorGraphTestUtil") Nothing 
      setTargets [t1]
      result <- load LoadAllTargets
      liftIO $ (succeeded result) @? "Couldn't load sample module"

      let builtInDataTyCons = [ (unitTyCon, [])
                              , (pairTyCon, tyConDataCons pairTyCon)
                              , (intTyCon, [])
                              , (listTyCon, tyConDataCons listTyCon)
                              ]
      extraDataTyCons <- mapM (\tq -> (id &&& tyConDataCons) <$> tyconFromTHName tq ) [''Diverging, ''Maybe]
      let tyConMap = mconcat . map (uncurry UniqMap.singleton . second (mconcat . map UniqSet.singleton)) $ (builtInDataTyCons ++ extraDataTyCons)
      testIdsWithTypes <- mapM (\tid -> (tid,) <$> typeFromTHTypeQ (testIdTypeQ tid)) testIds

      let cg = foldl' CG.addRoot (CG.empty tyConMap limit) testIdsWithTypes

      mapM_ ( $ cg) mkAssertions
  
  

notWrittenYet :: Maybe HUnit.Test
notWrittenYet = Nothing

gives :: TestId -> [(TestId, [CGExpected])] -> ConstructorGraph TestId -> Ghc ()
gives tid expChain cg = do
  --liftIO $ putStrLn (dotifyConstructorGraph cg)
  tyToLookup <- typeFromTHTypeQ . testIdTypeQ $ tid
  let actChain = CG.destructPlan tyToLookup cg
  liftIO $ (length actChain == length expChain) @? "Expected chain has different length to actual chain.  Got: " ++ (show . map fst $ actChain) ++ ", expected: " ++ show expChain ++ ", for : " ++ show tyToLookup

  forM_ expChain $ \(expId, expParts) -> do
    let [(actId, actDestruct)] = filter ((== (testIdId expId)) . testIdId . fst . fst) actChain
    checkDestruct expParts actDestruct

checkDestruct :: [CGExpected] -> DestructPlan -> Ghc ()
checkDestruct exps dPlan = do
  let dCMap = fromMaybe (error "DestructPlan didn't step when expected" ) 
                        (CG.step dPlan)
  liftIO $ length (concatMap IntMap.elems $ UniqMap.values dCMap) == length exps @? "DestructPlan stepped to incorrect number of alternatives. Got: " ++ (show $ map fst . UniqMap.elems $ dCMap) ++ ", expected: " ++ show exps

  forM_ exps (checkExpected dCMap)

checkExpected :: UniqMap DataCon (IntMap DestructPlan) -> CGExpected -> Ghc ()
checkExpected dCMap (CGE (thDCName, dcArg) rhs) = do
  dcExp <- dataconFromTHName thDCName
  let (Just nextDP) = IntMap.lookup dcArg (UniqMap.lookup dCMap dcExp)
  case rhs of
    StopsHere     -> liftIO $ isNothing (CG.step nextDP) @? "DestructPlan shouldn't have stepped when it did"
    CarrysOn      -> liftIO $ isJust (CG.step nextDP) @? "DestructPlan didn't step when it should have"
    Multiple exps -> checkDestruct exps nextDP
  

(==>) :: TestId -> [ CGExpected ] -> (TestId, [CGExpected])
(==>) = (,)


n :: CGRHS
n = StopsHere

(!) :: CGRHS
(!) = CarrysOn

t :: Int -> TH.TypeQ -> TestId
t n tq = TestId tq n
