{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE TupleSections #-}
module IdDB.TestLib where

import Control.Monad
import Control.Monad.Writer hiding (liftIO)
import GHC
import IdDB.Impl
import Language.Haskell.TH (TypeQ)
import MonadUtils
import qualified Language.Haskell.TH as TH
import qualified Util.UniqSet as UniqSet
import Test.HUnit ((~:), (@?))
import Type
import Util.GHC
import Util.TH
import Util.UniqSet (UniqSet)
import Var
import Data.Maybe

data There = There

data ExpectedProvider = TH.Name ::: TypeQ
                      | NoSubst TH.Name

newtype ConstraintM a = CM { runConstraintM :: Writer [Constraint] a }
  deriving Monad

data Constraint
  = Is TypeQ [ExpectedProvider]
  | IsNot TypeQ

data Test
  = Test
  { testNames       :: [TH.Name]
  , testConstraints :: [Constraint]
  , idSourceToUse   :: IdSource
  }

(-->) :: TypeQ -> TypeQ -> TypeQ
x --> y = liftM2 (\x y -> TH.AppT (TH.AppT TH.ArrowT x) y) x y

class Present a where
  getPresent :: a -> [ExpectedProvider]

instance Present TH.Name where
  getPresent name = [NoSubst name]

instance Present a => Present [a] where
  getPresent = concatMap getPresent

instance Present ExpectedProvider where
  getPresent x = [x]

there :: There
there = There

with :: [TH.Name] -> ConstraintM () -> Maybe Test
with names (CM (execWriter -> cs)) 
  = Just $ Test { testNames       = names
                , testConstraints = cs
                , idSourceToUse   = SourceModule
                }

is :: (Present ids) => TypeQ -> ids -> ConstraintM ()
is ty ep = CM $ tell [Is ty (getPresent ep)]

isn't :: TypeQ -> There -> ConstraintM ()
isn't tq There = CM $ tell [IsNot tq]

notWrittenYet :: Maybe Test
notWrittenYet = Nothing

mkExp :: ExpectedProvider -> Ghc (TvSubst -> IO (), Id)
mkExp (NoSubst nam) 
  = (\s -> isEmptyTvSubst s @? "Expected empty subst, got: " ++ show s,) <$> idFromTHName nam
mkExp (nam ::: tyq)   = do
  id <- idFromTHName nam
  ty <- dropForAlls <$> typeFromTHTypeQ tyq

  let rawType = dropForAlls . varType $ id
  let expF subst = isJust (unifyTypes st ty) @? "Expected a subst' type of: " ++ showDebug ty ++ ", got: " ++ showDebug st ++ ", from: " ++ showDebug subst
                    where 
                      st = substTy subst rawType 

  return $ (expF, id)

checkOne :: (TvSubst -> IO (), Id) -> (TvSubst, Id) -> IO ()
checkOne (es, eid) (as, aid) = do
  eid == aid @? "Expected id: " ++ show eid ++ ", got: " ++ show aid
  es as 

flatten :: [(TvSubst, UniqSet Id)] -> [(TvSubst, Id)]
flatten xs = do
  (subst, ids) <- xs
  id <- UniqSet.toList ids
  return $ (subst, id)

