{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE TupleSections #-}
module IdDB.TypeMapTestLib
( with
, is
, isn't
, there
, ExpectedProvider ((:::))
, (-->)
, notWrittenYet
, buildTests
) where

import Control.Monad
import Data.List
import Data.Maybe
import Data.Ord
import GHC
import IdDB.TypeMap (TypeMap)
import MonadUtils
import GHC.Paths
import qualified IdDB.TypeMap as TM
import qualified Test.HUnit as HUnit
import Test.HUnit ((~:), (@?))
import Type
import Util.TH
import IdDB.TestLib

buildTests :: (Int -> Maybe Test) -> HUnit.Test
buildTests f = HUnit.TestList $ map mkTest . catMaybes . takeWhile isJust . map (f$) $ [0..]

mkTest :: Test -> HUnit.Test
mkTest test = show (testNames test) ~: runGhc (Just libdir) $ do
  dflags <- getSessionDynFlags 
  setSessionDynFlags dflags
  defaultErrorHandler dflags $ do
    defaultCleanupHandler dflags $ do
      guessTarget ("IdDB.TypeMapSampleModule") Nothing >>= \target -> setTargets [target]
      result <- load LoadAllTargets
      liftIO $ (succeeded result) @? "Couldn't load sample module"


      names <- mapM idFromTHName (testNames test) 
      let typeMap = foldl TM.addProvider TM.empty names
      
      forM_ (testConstraints test) (checkConstraint typeMap)

checkConstraint :: TypeMap -> Constraint -> Ghc ()
checkConstraint tm (IsNot tyq) = do
  ty <- typeFromTHTypeQ tyq
  let found = TM.lookupProviders (dropForAlls ty) tm
  liftIO $ (null found) @? "Looking up " ++ show ty ++ ", expected: [], got: " ++ show found
checkConstraint tm (Is tyq thExps) = do
  ty <- typeFromTHTypeQ tyq
  exps :: [(TvSubst -> IO (), Id)] <- mapM mkExp thExps

  let found :: [(TvSubst, Id)] = flatten $ TM.lookupProviders (dropForAlls ty) tm
  let srt = sortBy (comparing snd)

  liftIO $ (length found == length exps) @? "Didn't get equal numbers of expected and found!  Expected: " ++ show (map snd exps) ++ ", got: " ++ show (map snd found)
  liftIO $ zipWithM_ checkOne (srt exps) (srt found)

