{-# LANGUAGE RankNTypes #-}
{-# LANGUAGE TemplateHaskell #-}
module Plan.PlanTest where

{- original plans with old IDDB -}
import qualified Plan.Base0 as Base0
import qualified Plan.Base1 as Base1
import qualified Plan.Base2 as Base2
import qualified Plan.Base3 as Base3
import qualified Plan.Base4 as Base4
import qualified Plan.HasFailCase as HasFailCase
import qualified Plan.HasFunctionTarget as HasFunctionTarget
import qualified Plan.LoopExample as LoopExample
import qualified Plan.LoopExample2 as LoopExample2
import qualified Plan.Trivial as Trivial
import qualified Plan.Trivial1 as Trivial1
import qualified Plan.DelayedMinimalEdge as DelayedMinimalEdge
import qualified Plan.NoDepth as NoDepth
import qualified Plan.PolymorphicFree as PolymorphicFree
import qualified Plan.PolymorphicInstantiate as PolymorphicInstantiate
import qualified Plan.PolymorphicConstraint as PolymorphicConstraint
import qualified Plan.PolymorphicChaseConstraint as PolymorphicChaseConstraint
import qualified Plan.PolymorphicChaseConstraintSmall as PolymorphicChaseConstraintSmall
import qualified Plan.PolymorphicSuperfluous as PolymorphicSuperfluous
import qualified Plan.CanonicalBottom as CanonicalBottom
import qualified Plan.ListTest as ListTest
import qualified Plan.UnifyLatter as UnifyLatter
import qualified Plan.UnifyHO as UnifyHO
import qualified Plan.UnifyHO2 as UnifyHO2
import qualified Plan.Types as Types

{- new plans with new IdDB -}
import qualified Plan.UnifyN as UnifyN
import qualified Plan.UnifyDTN as UnifyDTN
import qualified Plan.UnifyDTSimpleN as UnifyDTSimpleN
import qualified Plan.ListTestN as ListTestN
import qualified Plan.MaybeTheN as MaybeTheN
import qualified Plan.CaseList1 as CaseList1
import qualified Plan.CaseList2 as CaseList2
import qualified Plan.CaseTuple1 as CaseTuple1
import qualified Plan.CaseTuple2 as CaseTuple2
import qualified Plan.CaseTuple3 as CaseTuple3

import qualified Plan.UnifyNested as UnifyNested
import qualified Plan.NestedCalls as NestedCalls
import qualified Plan.CaseForArgument as CaseForArgument
import qualified Plan.CaseForArgument2 as CaseForArgument2
import qualified Plan.Board as Board

import IdDB (IdSource(..))
import Plan.TestLib
import Plan.Types
import Test.HUnit ((~:), (@?))

alpha     = [t|Alpha|]
beta      = [t|Beta|]
gamma     = [t|Gamma|]
delta     = [t|Delta|]
person    = [t|Person|]
name      = [t|Name|]
bool      = [t|Bool|]
string    = [t|String|]
carryNil  = [t|Carry Nil|]
carryBool = [t|Carry Bool|]
holdBool  = [t|Holder Bool|]
holdInt   = [t|Holder Int|]
nil       = [t|Nil|]
unit      = [t|()|]

tests = "PlanTest" ~: ((map mkPlanTest allTests) ++ ["TODOs" ~: (False @? "TODO - tests for different discovery strategies")])

allTests = takeWhile isATest . map t $ [0..]

t 0 = test { modu    = "Base0"
           , rootFunction = 'Base0.f
           , expectedIdsForTypes = mkNoSubstEIFT [ alpha --> ['Base0.f] ]
           }

t 1 = test { modu         = "Base1"
           , rootFunction = 'Base1.f
           , expectedIdsForTypes = mkNoSubstEIFT
                                   [ beta  --> ['Base1.f]
                                   , alpha --> ['Base1.g]
                                   , (alpha |> beta) --> ['Base1.f]
                                   ]
           }

t 2 = test { modu    = "Base2"
           , rootFunction  = 'Base2.f
           , expectedIdsForTypes = mkNoSubstEIFT
                                   [ alpha --> ['Base2.f]
                                   , (gamma |> alpha) --> ['Base2.f]
                                   , (beta |> gamma |> alpha) --> ['Base2.f]
                                   , gamma --> ['Base2.g]
                                   , (beta |> gamma) --> ['Base2.g]
                                   , beta  --> ['Base2.h]
                                   ]
           , treePlanMaxDepth = 10
           }

t 3 = test { modu   = "Base3"
           , rootFunction = 'Base3.f
           , expectedIdsForTypes = mkNoSubstEIFT
                                   [ alpha --> ['Base3.f]
                                   , (gamma |> alpha) --> ['Base3.f]
                                   , gamma --> ['Base3.g, 'Base3.h]
                                   , (gamma |> gamma) --> ['Base3.g]
                                   ]
           , treePlanMaxDepth = 16
           }

t 4 = test { modu   = "HasFailCase"
           , rootFunction = 'HasFailCase.f
           , expectedIdsForTypes = mkNoSubstEIFT
                                   [ alpha --> ['HasFailCase.f]
                                   , (beta |> alpha) --> ['HasFailCase.f]
                                   , (delta |> beta) --> ['HasFailCase.g]
                                   , beta --> ['HasFailCase.g, 'HasFailCase.h]
                                   ]
           , treePlanMaxDepth = 10
           }

t 5 = test { modu   = "HasFunctionTarget"
           , rootFunction = 'HasFunctionTarget.f
           , expectedIdsForTypes = mkNoSubstEIFT
                                   [ alpha --> ['HasFunctionTarget.f]
                                   , ((alpha |> beta) |> alpha) --> ['HasFunctionTarget.f]
                                   , beta --> ['HasFunctionTarget.g]
                                   , (alpha |> beta) --> ['HasFunctionTarget.g]
                                   ]
           }

t 6 = test { modu = "Trivial"
           , rootFunction = 'Trivial.test
           , expectedIdsForTypes = mkNoSubstEIFT
                                   [ bool --> ['Trivial.test]
                                   , (person |> bool) --> ['Trivial.test]
                                   , person --> ['Trivial.julia, 'Trivial.carla]
                                   ]
           }

t 7 = test { modu = "Trivial1"
           , rootFunction = 'Trivial1.test
           , expectedIdsForTypes = mkNoSubstEIFT
                                   [ bool --> ['Trivial1.test]
                                   , (person |> bool) --> ['Trivial1.test]
                                   , name --> ['Trivial1.julia, 'Trivial1.carla]
                                   , person --> ['Trivial1.mkPerson]
                                   , (name |> person) --> ['Trivial1.mkPerson]
                                   , string --> ['Trivial1.debugPerson]
                                   , (person |> string) --> ['Trivial1.debugPerson]
                                   ]
           , treePlanMaxDepth = 10
           }

t 8 = test { modu = "LoopExample"
           , rootFunction = 'LoopExample.test
           , expectedIdsForTypes = mkNoSubstEIFT
                                   [ bool --> ['LoopExample.test]
                                   , (person |> bool) --> ['LoopExample.test]
                                   , person --> ['LoopExample.makeTaller, 'LoopExample.makePerson]
                                   , (person |> person) --> ['LoopExample.makeTaller]
                                   , (name |> person) --> ['LoopExample.makePerson]
                                   , name --> ['LoopExample.julia]
                                   ]
           , treePlanMaxDepth = 12
           }

t 9 = test { modu   = "LoopExample2"
           , rootFunction = 'LoopExample2.test
           , expectedIdsForTypes = mkNoSubstEIFT
                                   [ bool --> ['LoopExample2.test]
                                   , (person |> bool) --> ['LoopExample2.test]
                                   , name --> ['LoopExample2.getName, 'LoopExample2.julia]
                                   , (person |> name) --> ['LoopExample2.getName]
                                   , (name |> person) --> ['LoopExample2.mkPerson]
                                   , person --> ['LoopExample2.mkPerson]
                                   ]
           , treePlanMaxDepth = 12
           }

t 10 = test { modu   = "DelayedMinimalEdge"
            , rootFunction = 'DelayedMinimalEdge.f
            , expectedIdsForTypes = mkNoSubstEIFT
                                    [ alpha --> ['DelayedMinimalEdge.a]
                                    , delta --> ['DelayedMinimalEdge.f, 'DelayedMinimalEdge.g]
                                    , (alpha |> delta) --> ['DelayedMinimalEdge.g]
                                    , (alpha |> alpha |> delta) --> ['DelayedMinimalEdge.g]
                                    , (alpha |> alpha |> alpha |> delta) --> ['DelayedMinimalEdge.g]
                                    , (alpha |> alpha |> alpha |> alpha |> delta) --> ['DelayedMinimalEdge.g]
                                    , ((alpha |> alpha |> delta) |> delta) --> ['DelayedMinimalEdge.f]
                                    , ((alpha |> alpha |> alpha |> delta) |> (alpha |> alpha |> delta) |> delta) --> ['DelayedMinimalEdge.f]
                                    ]
            , treePlanMaxDepth = 10
            }

t 11 = test { modu    = "NoDepth"
            , rootFunction = 'NoDepth.f
            , expectedIdsForTypes = mkNoSubstEIFT
                                    [alpha --> ['NoDepth.f]
                                    ]
            , treePlanMaxDepth = 0
            }

t 12 = test { modu = "PolymorphicFree"
            , rootFunction = 'PolymorphicFree.test
            , expectedIdsForTypes = [ {- URK, can't actually test this bool --> [('PolymorphicFree.test, Just ([t| forall a . a -> Bool |]))]
                                    ,  -} (bool |> bool) --> [('PolymorphicFree.test, Just (bool |> bool))] 
                                       -- TODO Think this should be []
                                    ]
            , treePlanMaxDepth = 5
            }

t 13 = test { modu = "PolymorphicInstantiate"
            , rootFunction = 'PolymorphicInstantiate.test
            , expectedIdsForTypes = [ bool --> [('PolymorphicInstantiate.test, Nothing)]
                                    , nil  --> [('PolymorphicInstantiate.nil, Nothing)]
                                    , carryNil --> [('PolymorphicInstantiate.mkCarry, Just (nil |> carryNil))]
                                    , (carryNil |> bool) --> [('PolymorphicInstantiate.test, Nothing)]
                                    , (bool |> carryBool) --> [('PolymorphicInstantiate.mkCarry, Just (bool |> carryBool))]
                                    ]
            , treePlanMaxDepth = 10
            }

t 14 = test { modu = "PolymorphicConstraint"
            , rootFunction = 'PolymorphicConstraint.test
            , expectedIdsForTypes = [ {- bool --> [('PolymorphicConstraint.test, (Just ([t| forall a . Carry a -> a -> Bool |])))]  -- URK need to test this too!
                                    , -} carryNil --> [('PolymorphicConstraint.mkCarry, Nothing)]
                                    , (bool |> bool) --> [('PolymorphicConstraint.test, Just (carryBool |> bool |> bool))]  -- TODO is this sensible? Should I worry about it later?
                                    , (carryBool |> bool |> bool) --> [('PolymorphicConstraint.test, Just (carryBool |> bool |> bool))]
                                    , (carryNil |> nil |> bool) --> [('PolymorphicConstraint.test, Just (carryNil |> nil |> bool))]
                                    ]
            , treePlanMaxDepth = 10
            }

t 15 = test { modu = "PolymorphicSuperfluous"
            , rootFunction = 'PolymorphicSuperfluous.test
            , expectedIdsForTypes = mkNoSubstEIFT
                                    [ nil --> ['PolymorphicSuperfluous.nil]
                                    , bool --> ['PolymorphicSuperfluous.test]
                                    , (nil |> bool) --> ['PolymorphicSuperfluous.test]
                                    , (bool |> bool) --> [] -- TODO is this sensible?
                                    , (string |> string) --> []
                                    ]
            , treePlanMaxDepth = 10
            }

t 16 = test { modu = "PolymorphicChaseConstraint"
            , rootFunction = 'PolymorphicChaseConstraint.test
            , expectedIdsForTypes = [] -- Just want to test the Plan component itself
            , treePlanMaxDepth = 10
            }


t 17 = test { modu = "PolymorphicChaseConstraintSmall"
            , rootFunction = 'PolymorphicChaseConstraintSmall.test
            , expectedIdsForTypes = [] -- Just want to test the Plan component itself
            , treePlanMaxDepth = 10
            }


t 18 = test { modu = "CanonicalBottom"
            , rootFunction = 'CanonicalBottom.test
            , expectedIdsForTypes = [] -- Just want to test the Plan component itself
            , treePlanMaxDepth = 10
            }

t 19 = test { modu = "ListTest"
            , rootFunction = 'ListTest.test
            , expectedIdsForTypes = [] -- Just want to test the Plan component itself
            , treePlanMaxDepth = 10
            }

t 20 = test { modu = "UnifyLatter"
            , rootFunction = 'UnifyLatter.test
            , expectedIdsForTypes = [] -- Just want to test the Plan component itself
            , treePlanMaxDepth = 10
            }

{- TODO this is knocked out as it's too big -}
t 21 = test { modu = "UnifyHO"
            , rootFunction = 'UnifyHO.test
            , expectedIdsForTypes = [] -- Just want to test the Plan component itself
            , treePlanMaxDepth = 10
            }

t 22 = test { modu = "UnifyHO2"
            , rootFunction = 'UnifyHO2.test
            , expectedIdsForTypes = [] -- Just want to test the Plan component itself
            , treePlanMaxDepth = 6
            }

t 23 = test { modu = "UnifyN"
            , rootFunction = 'UnifyN.test
            , expectedIdsForTypes = []
            , treePlanMaxDepth = 4
            , idSourceToUse = SourceModule
            }

t 24 = test { modu = "UnifyDTN"
            , rootFunction = 'UnifyDTN.test
            , expectedIdsForTypes = [ [t|Maybe Int|] --> [ ('Just, Just [t|Int -> Maybe Int|])
                                                         , ('Nothing, Just [t|Maybe Int|])
                                                         ]
                                    , [t|forall z . Maybe z|] --> [('Just, Just [t| forall z . z -> Maybe z|])
                                                                  ,('Nothing, Just [t| forall z . Maybe z |])
                                                                  ]
                                    ]
            , idSourceToUse = SourceModule
            }

t 25 = test { modu = "UnifyDTSimpleN"
            , rootFunction = 'UnifyDTSimpleN.test
            , expectedIdsForTypes = []
            , idSourceToUse = SourceModule
            }

t 26 = test { modu = "ListTestN"
            , rootFunction = 'ListTestN.test
            , expectedIdsForTypes = [] -- Just want to test the Plan component itself
            , idSourceToUse = SourceModule
            }

t 27 = test { modu = "MaybeTheN"
            , rootFunction = 'MaybeTheN.test
            , expectedIdsForTypes = []
            , idSourceToUse = SourceModule
            }

t 28 = test { modu = "Base4"
            , rootFunction = 'Base4.test
            , expectedIdsForTypes = []
            , treePlanMaxDepth = 10
            }

t 29 = test { modu = "CaseList1"
            , rootFunction = 'CaseList1.test
            , treePlanMaxDepth = 8
            , expectedIdsForTypes = []
            , treePlanCaseDestruct = True
            }

t 30 = test { modu = "CaseList2"
            , rootFunction = 'CaseList2.testNested
            , treePlanMaxDepth = 8
            , expectedIdsForTypes = []
            , treePlanCaseDestruct = True
            }

t 31 = test { modu = "CaseTuple1"
            , rootFunction = 'CaseTuple1.test
            , treePlanMaxDepth = 8
            , expectedIdsForTypes = []
            , treePlanCaseDestruct = True
            }

t 32 = test { modu = "CaseTuple2"
            , rootFunction = 'CaseTuple2.test
            , treePlanMaxDepth = 9
            , expectedIdsForTypes = []
            , treePlanCaseDestruct = True
            }

t 33 = test { modu = "CaseTuple3"
            , rootFunction = 'CaseTuple3.test
            , treePlanMaxDepth = 10
            , expectedIdsForTypes = []
            , treePlanCaseDestruct = True
            }

t 34 = test { modu = "UnifyNested"
            , rootFunction = 'UnifyNested.test
            , treePlanMaxDepth = 6
            , expectedIdsForTypes = []
            , idSourceToUse = SourceModule
            }

t 35 = test { modu = "NestedCalls"
            , rootFunction = 'NestedCalls.test
            , treePlanMaxDepth = 15
            , expectedIdsForTypes = []
            , idSourceToUse = TestModule
            }

t 36 = test { modu = "CaseForArgument"
            , rootFunction = 'CaseForArgument.prop_Sad
            , treePlanMaxDepth = 13
            , expectedIdsForTypes = []
            , idSourceToUse = SourceModule
            , treePlanCaseArgument = True
            }

t 37 = test { modu = "CaseForArgument2"
            , rootFunction = 'CaseForArgument2.prop_Sad
            , treePlanMaxDepth = 13
            , expectedIdsForTypes = []
            , idSourceToUse = SourceModule
            , treePlanCaseArgument = True
            }

t 38 = test { modu = "Board"
            , rootFunction = 'Board.step
            , treePlanMaxDepth = 10
            , expectedIdsForTypes = []
            , idSourceToUse = SourceModule
            , treePlanCaseArgument = True
            }

t _ = noTestYet
