{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE MultiParamTypeClasses #-}
module SCA.SimpleTestLib
( buildTests
, elem
, noConstants
, notWrittenYet
, with
, lStringConstants
, i
) where

import Prelude hiding (elem)
import qualified Language.Haskell.TH as TH
import Test.HUnit ((~:), (@?))
import qualified Test.HUnit as HUnit

import SCA.Simple
import Data.Maybe
import Control.Monad.Writer.Strict hiding (liftIO)
import Util.FCLabels
import qualified Data.Foldable as F
import Control.Applicative
import FastString
import qualified Util.UniqSet as UniqSet
import qualified Data.Set as Set
import Data.Set (Set)
import Util.UniqSet (UniqSet)

import GHC
import GHC.Paths
import Digraph
import Control.Applicative
import Util.TH
import MonadUtils
import HscTypes
import CoreSyn
import Util.GHC
import Outputable

buildTests :: (Int -> Maybe Test) -> HUnit.Test
buildTests f = HUnit.TestList $ map mkTest . catMaybes . takeWhile isJust . map (f$) $ [0..]

mkTest :: Test -> HUnit.Test
mkTest test = show (testName test) ~: do
  runGhc (Just libdir) $ do
    dflags <- getSessionDynFlags 
    setSessionDynFlags dflags

    defaultErrorHandler dflags $ do
      defaultCleanupHandler dflags $ do
        guessTarget "SCA.TestFunctions" Nothing >>= \target -> setTargets [target]
        load LoadAllTargets 
        mg <- depanal [] False
        let mods = flattenSCCs (topSortModuleGraph False mg Nothing)
        let [tfSummary] = filter ((== "SCA.TestFunctions") . moduleNameString . ms_mod_name) mods

        modGuts <- coreModule <$> (desugarModule =<< typecheckModule =<< parseModule tfSummary)
        testId <- idFromTHName (testName test)

        let Just expr = lookup testId . flattenBinds . mg_binds $ modGuts

        let actualConstants   = findConstants expr 
        let expectedConstants = testConstants test

        liftIO $ expectedConstants == actualConstants @? "Expected constants not equalt to actual constants"

        return ()



data Test
  = Test 
  { testName      :: TH.Name
  , testConstants :: Constants
  }

with :: TH.Name -> ConstraintM () -> Maybe Test
with n (CM (execWriter -> cs)) 
  = Just $ Test { testName      = n
                , testConstants = cs
                }

notWrittenYet :: Maybe Test
notWrittenYet = Nothing

newtype ConstraintM a = CM { runConstraintM :: Writer Constants a }
  deriving Monad

noConstants :: ConstraintM ()
noConstants = return ()

class ElemPair a b where
  getConstants :: a -> b -> Constants

elem :: ElemPair a b => a -> b -> ConstraintM ()
elem a b = CM (tell (getConstants a b))

instance ElemPair a (Constants :-> Set a) where
  getConstants e f = set f (Set.singleton e) mempty

instance (Ord a) => ElemPair [a] (Constants :-> (Set a)) where
  getConstants es f = set f (Set.fromList es) mempty

i :: Integer -> Integer
i = id

lStringConstants :: Constants :-> Set String
lStringConstants = label (Set.fromList . map unpackFS . UniqSet.toList . get lFastStringConstants) 
                         (set lFastStringConstants . (F.foldMap (UniqSet.singleton . mkFastString) . Set.toList))
